import warnings

import psutil
from rich import get_console
from rich.columns import Columns
from rich.live import Live
from rich.spinner import Spinner
from rich.table import Table


def _pretty_warning(msg, *args, **kwargs):
    return Colors.str(f"WARNING: {msg}\n", properties=["warning"])


warnings.formatwarning = _pretty_warning


class Colors:
    SUCCESS = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    BLUE = "\033[94m"
    BOLD = "\033[1m"
    _ENDC = "\033[0m"

    @staticmethod
    def _props2str(props):
        return "".join([getattr(Colors, prop.upper()) for prop in props])

    @staticmethod
    def str(string, properties=[]):
        return Colors._props2str(properties) + string + Colors._ENDC

    @staticmethod
    def print(*args, properties=[], **kwargs):
        print(Colors._props2str(properties), end="")
        print_end = kwargs.pop("end", "\n")
        print(*args, **kwargs, end=Colors._ENDC + print_end)

    @staticmethod
    def warn(*args, **kwargs):
        warnings.warn(*args, **kwargs)


def print_table(columns, rows, justifications=["left", "center"]):
    table = Table(show_lines=True)
    for i, column in enumerate(columns):
        table.add_column(column, justify=justifications[i])

    for row in rows:
        row = list(map(str, row))
        max_lines = max(string.count("\n") for string in row)
        vert_align_row = [
            "\n" * int((max_lines - string.count("\n")) / 2) + string for string in row
        ]
        table.add_row(*vert_align_row)

    with get_console() as console:
        console.print(table)


def count_cores():
    cores = psutil.cpu_count(logical=False)
    print("Cores:", cores)
    threads = psutil.cpu_count(logical=True) / cores
    int_threads = int(threads)
    if int_threads == threads:
        print("Threads per core:", int_threads)


def warnings_suppressed():
    return any(
        [filter[0] == "ignore" and filter[2] is Warning for filter in warnings.filters]
    )


def loading(msg):
    loading_text = Columns([msg, Spinner("simpleDotsScrolling")])
    return Live(loading_text, refresh_per_second=5, transient=True)


def yes_or_no(prompt):
    while True:
        response = input(prompt).lower().strip()

        if response == "":
            continue

        responded_yes = response == "yes"[: len(response)]
        responded_no = response == "no"[: len(response)]

        if responded_yes != responded_no:  # responeded_yes xor responded_no
            return responded_yes
        else:
            Colors.print(
                'Invalid response. Please enter either "y" or "n"', properties=["fail"]
            )


def validate_num(val, dtype):
    try:
        cast_val = dtype(val)
    except ValueError:
        return

    if dtype is int and cast_val != float(val):
        return
    return cast_val



import builtins
import os
import signal
import sys
import time
from multiprocessing import Pool
from multiprocessing.managers import BaseManager
from multiprocessing.pool import ThreadPool
from threading import Event, Lock

import psutil
from bs4 import BeautifulSoup
from requests import Session
from requests.adapters import HTTPAdapter
from requests.exceptions import RequestException
from requests.packages.urllib3.util.retry import Retry
from selenium.common.exceptions import TimeoutException
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait

from best_buy_bullet_bot.audio import sound_effects
from best_buy_bullet_bot.browser import purchase
from best_buy_bullet_bot.data import user_data
from best_buy_bullet_bot.data.setting_utils import (
    DRIVER_NAMES,
    SOUND_MODES,
    MoneyManager,
    get_settings,
)
from best_buy_bullet_bot.data.url_utils import QtyManager
from best_buy_bullet_bot.tracker.progress_bar import IndefeniteProgressBar
from best_buy_bullet_bot.utils import Colors

WINDOWS = sys.platform == "win32"
SETTINGS = get_settings()
SOUND_MODE = SETTINGS["sound mode"]
AUTO_CHECKOUT = SETTINGS["auto checkout"]
BROWSER_NAME = SETTINGS["browser"]
DRIVER_WRAPPER = DRIVER_NAMES[BROWSER_NAME]
NUM_THREADS = SETTINGS["threads"]


class TwoWayPause:
    def __init__(self):
        self.play = Event()
        self.play.set()
        self.pause = Event()

    def is_set(self):
        return self.pause.is_set()

    def set(self):
        self.play.clear()
        self.pause.set()

    def clear(self):
        self.pause.clear()
        self.play.set()

    def wait(self):
        self.pause.wait()

    def wait_inverse(self):
        self.play.wait()


# The following two classes are needed to use magic methods with `BaseManager`
class NoMagicLock:
    def __init__(self):
        self.lock = Lock()

    def enter(self, *args, **kwargs):
        self.lock.__enter__(*args, **kwargs)

    def exit(self, *args, **kwargs):
        self.lock.__exit__(*args, **kwargs)


class NormalLock:
    def __init__(self, no_magic_lock):
        self.lock = no_magic_lock

    def __enter__(self, *args, **kwargs):
        self.lock.enter(*args, **kwargs)

    def __exit__(self, *args, **kwargs):
        self.lock.exit(*args, **kwargs)


BaseManager.register("ThreadLock", NoMagicLock)
BaseManager.register("QtyManager", QtyManager)
BaseManager.register("IndefeniteProgressBar", IndefeniteProgressBar)
BaseManager.register("PauseEvent", TwoWayPause)
BaseManager.register("MoneyManager", MoneyManager)


STOCK = False


def track(
    title,
    url,
    qty,
    headless,
    login_cookies,
    password,
    cvv,
    thread_lock,
    paused,
    pbar,
    pred_price,
    money_manager,
    headers,
):
    builtins.print = pbar.print
    if not AUTO_CHECKOUT:
        headers["referer"] = url

    thread_lock = NormalLock(thread_lock)

    with ThreadPool(NUM_THREADS) as pool:
        pool.starmap_async(
            run,
            [
                [
                    title,
                    url,
                    qty,
                    login_cookies,
                    thread_lock,
                    paused,
                    pbar,
                    pred_price,
                    money_manager,
                    headers,
                ]
                for _ in range(NUM_THREADS)
            ],
        )

        await_checkout(
            title,
            url,
            qty,
            headless,
            login_cookies,
            password,
            cvv,
            paused,
            pred_price,
            money_manager,
        )


def await_checkout(
    title,
    url,
    qty,
    headless,
    login_cookies,
    password,
    cvv,
    paused,
    pred_price,
    money_manager,
):
    global STOCK

    while True:
        paused.wait()

        if STOCK:
            if not money_manager.check_funds(pred_price) or not qty.get():
                paused.clear()

                if SOUND_MODE == SOUND_MODES[2]:
                    sound_effects.stop()

                Colors.print(
                    f'All requested "{title}" were purchased.'
                    if money_manager.check_funds(pred_price)
                    else f"With only ${money_manager.get_funds():,.2f} you cannot afford {title}.",
                    "It will no longer be tracked to conserve resources.\n",
                    properties=["warning"],
                )
                return

            current_time = time.strftime("%H:%M:%S", time.localtime())
            Colors.print(
                f'\n{current_time} - "{title}" - {url}\n',
                properties=["bold"],
            )

            # Plays a sound
            if SOUND_MODE == SOUND_MODES[1]:
                sound_effects.play()
            elif SOUND_MODE == SOUND_MODES[2]:
                sound_effects.start()

            if AUTO_CHECKOUT:
                try:
                    while money_manager.check_funds(pred_price) and qty.get():
                        if purchase(
                            url,
                            login_cookies,
                            headless,
                            *STOCK,  # `headless_driver` and `headless_wait`
                            title,
                            password,
                            cvv,
                            money_manager,
                        ):
                            qty.decrement()
                        else:
                            break
                except Exception as e:
                    Colors.print(
                        f"CHECKOUT ERROR: {e}",
                    )

            STOCK = False
            paused.clear()
        else:
            paused.wait_inverse()


def run(
    title,
    url,
    qty,
    login_cookies,
    thread_lock,
    paused,
    pbar,
    pred_price,
    money_manager,
    headers,
):
    global STOCK

    stop_tracker = False

    if AUTO_CHECKOUT:
        options = DRIVER_WRAPPER.options()
        options.page_load_strategy = "none"
        options.add_argument("--proxy-server='direct://'")
        options.add_argument("--proxy-bypass-list=*")
        options.add_argument("--headless")

        # Suppress "DevTools listening on ws:..." message
        if BROWSER_NAME == "chrome":
            options.add_experimental_option("excludeSwitches", ["enable-logging"])

        # Create the browser window
        driver = DRIVER_WRAPPER.driver(
            executable_path=DRIVER_WRAPPER.manager().install(), options=options
        )

        # Login to the browser by setting the cookies
        driver.get(url)
        for cookie in login_cookies:
            driver.add_cookie(cookie)

        wait = WebDriverWait(driver, 120)
        button_locator = EC.presence_of_element_located(
            (
                By.CSS_SELECTOR,
                ".fulfillment-add-to-cart-button > div > div > button",
            )
        )

        # Confirm that we have a stable connection and that Best Buy hasn't made any
        # changes to their website that would break out locator
        try:
            btn = wait.until(button_locator)
        except TimeoutException:
            Colors.print(
                f"Unable to connect to {title}. Closing tracker.",
                properties=["fail"],
            )
            stop_tracker = True
    else:
        session = Session()
        session.headers.update(headers)
        retry = Retry(
            connect=3,
            backoff_factor=0.25,
            status_forcelist=[429, 500, 502, 503, 504],
            method_whitelist=["HEAD", "GET", "OPTIONS"],
        )
        adapter = HTTPAdapter(max_retries=retry)
        session.mount("https://", adapter)
        session.mount("http://", adapter)

    connection_status = True
    available = False
    prev_available = False

    # Track item so long as we have sufficient funds and haven't bought the item too many times
    while not stop_tracker and (
        not AUTO_CHECKOUT or (money_manager.check_funds(pred_price) and qty.get())
    ):
        # Stop trackers to conserve resources during the auto checkout process
        if paused.is_set():
            paused.wait_inverse()
            continue

        if AUTO_CHECKOUT:
            driver.get(url)

            try:
                # Wait until old page has unloaded
                wait.until(EC.staleness_of(btn))

                if paused.is_set():
                    # Stop page load (page will reload when tracker restarts)
                    driver.execute_script("window.stop();")
                    continue

                # Wait until the add-to-cart button is present on the new page
                btn = wait.until(button_locator)

            # Inform the user if an error occurs while trying to locate the add-to-cart button
            except TimeoutException:
                if connection_status:
                    start_time = time.time()
                    Colors.print(
                        f"{title} tracker has lost connection.\n",
                        properties=["fail"],
                    )
                    connection_status = False
                continue

            # Check if it is an add-to-cart button
            available = btn.get_attribute("data-button-state") == "ADD_TO_CART"

        else:
            try:
                # Make a get request
                response = session.get(url, timeout=10)
                response.raise_for_status()

            except RequestException as e:
                # Inform the user if an error occurs while trying to make a get request
                if connection_status:
                    start_time = time.time()
                    Colors.print(
                        f"Unable to establish a connection to {title} remote endpoint.",
                        properties=["fail"],
                    )
                    print(e, "\n")
                    connection_status = False
                continue

            if paused.is_set():
                continue

            # Look for add-to-cart button
            soup = BeautifulSoup(response.text, "html.parser")
            available = (
                soup.find(
                    "button",
                    {"data-button-state": "ADD_TO_CART"},
                )
                is not None
            )

        pbar.update()

        # If we reconnected, inform the user
        if not connection_status:
            Colors.print(
                f"{title} tracker has successfully reconnected!",
                properties=["success"],
            )
            print(f"Downtime: {time.time()-start_time:.2f} seconds \n")
            connection_status = True

        # If the item is in stock
        if available:
            # Unlock the checkout process if it hasn't been already
            with thread_lock:
                if paused.is_set():
                    continue

                if AUTO_CHECKOUT:
                    if not STOCK:
                        STOCK = (driver, wait)
                else:
                    STOCK = True
                paused.set()

        # If item went back to being out of stock
        elif prev_available != available:
            if SOUND_MODE == SOUND_MODES[2]:
                sound_effects.stop()

        prev_available = available

    # Stop the auto checkout function
    if STOCK is not True:
        STOCK = True
        paused.set()

    if AUTO_CHECKOUT:
        driver.close()
    else:
        session.close()


def set_priority(high_priority):
    p = psutil.Process(os.getpid())

    """
    EASTER EGG: Thank you for reading the source code!
    To run the bot with a higher priority level and achieve better performance complete the following.

    If using Firefox, complete the following before moving on to the next step:
    WINDOWS: Open a Command Prompt window with "Run as administrator" https://www.educative.io/edpresso/how-to-run-cmd-as-an-administrator
    MAC: Enter the command `su` in your terminal to gain root privileges. Beware your settings may be different in the root session, but you can always return to a normal session with the `exit` command.

    Then regardless of your browser:
    Run `3b-bot --fast` in your shell.
    """
    # Windows: REALTIME_PRIORITY_CLASS, HIGH_PRIORITY_CLASS, ABOVE_NORMAL_PRIORITY_CLASS, NORMAL_PRIORITY_CLASS, BELOW_NORMAL_PRIORITY_CLASS, IDLE_PRIORITY_CLASS
    # MacOS: -20 is highest priority while 20 is lowest priority
    # Lower priorities are used here so other things can still be done on the computer while the bot is running
    priority = (
        (psutil.HIGH_PRIORITY_CLASS if WINDOWS else -10)
        if high_priority
        else (psutil.BELOW_NORMAL_PRIORITY_CLASS if WINDOWS else 10)
    )
    p.nice(priority)


def start(fast=False, headless=False, verify_account=False, skip_verification=False):
    from elevate import elevate

    from best_buy_bullet_bot.browser import browser_startup, get_user_agent
    from best_buy_bullet_bot.data import close_data, url_utils
    from best_buy_bullet_bot.utils import loading, warnings_suppressed, yes_or_no

    """
    EASTER EGG: Thank you for reading the source code!
    To run the bot with a higher priority level and achieve better performance complete the following.

    If using Firefox, complete the following before moving on to the next step:
    WINDOWS: Open a Command Prompt window with "Run as administrator" https://www.educative.io/edpresso/how-to-run-cmd-as-an-administrator
    MAC: Enter the command `su` in your terminal to gain root privileges. Beware your settings may be different in the root session, but you can always return to a normal session with the `exit` command.

    Then regardless of your browser:
    Run `3b-bot --fast` in your shell.
    """

    # If we don't have admin privileges try to elevate permissions
    if fast and hasattr(os, "getuid") and os.getuid() != 0:
        print("Elevating permissions to run in fast mode.")
        elevate(graphical=False)

    def kill_all(*args, **kwargs):
        if not WINDOWS:
            # Delete the temp file
            close_data()

        # Forcefully close everything
        os.system(
            f"taskkill /F /im {psutil.Process(os.getpid()).name()}"
        ) if WINDOWS else os.killpg(os.getpgid(os.getpid()), signal.SIGKILL)

    def clean_kill(*args, **kwargs):
        # Suppress error messages created as a result of termination
        # Selenium will often print long error messages during termination
        sys.stderr = open(os.devnull, "w")

        # Close the pbar and kill everything if the process pool has been created
        if "pbar" in locals():
            pbar.close()
            print()
            kill_all()  # Inelegant but fast

        # Otherwise we can exit the traditional way
        else:
            print()
            sys.exit(0)

    # Use custom functions to exit properly
    for sig in [signal.SIGINT, signal.SIGBREAK if WINDOWS else signal.SIGQUIT]:
        signal.signal(sig, clean_kill)
    signal.signal(signal.SIGTERM, kill_all)

    print(
        """
 .d8888b.  888888b.        888888b.            888
d88P  Y88b 888  "88b       888  "88b           888
     .d88P 888  .88P       888  .88P           888
    8888"  8888888K.       8888888K.   .d88b.  888888
     "Y8b. 888  "Y88b      888  "Y88b d88""88b 888
888    888 888    888      888    888 888  888 888
Y88b  d88P 888   d88P      888   d88P Y88..88P Y88b.
 "Y8888P"  8888888P"       8888888P"   "Y88P"   "Y888

"""
    )

    # Check if a flag has been passed to suppress warnings
    suppress_warnings = warnings_suppressed()

    raw_urls = url_utils.get_url_data()
    if not len(raw_urls):
        print()
        Colors.warn("No URLs have been set to be tracked.")
        if not suppress_warnings:
            if yes_or_no("Would you like to set some URLs for tracking (y/n): "):
                while True:
                    url_utils.add_url()
                    if not yes_or_no("Do you want to add another url (y/n): "):
                        break
                raw_urls = url_utils.get_url_data()

        if not len(raw_urls):
            Colors.print(
                "Not enough URLs for tracking.",
                "Please add at least 1 URL.",
                "URLs can be added with `3b-bot add-url`.",
                properties=["fail", "bold"],
            )
            sys.exit(1)

    print("Tracking the following URLs.")
    url_utils.view_urls(AUTO_CHECKOUT)

    manager = BaseManager()
    manager.start()

    money_manager = manager.MoneyManager()
    if AUTO_CHECKOUT:
        print(f"Current funds: ${money_manager.get_funds():,.2f}")
    print()

    # Get URLs and a quantity object for each URL
    urls, qtys = [], []
    for url_group, raw_qty in raw_urls:
        int_qty = -1 if raw_qty == "inf" else raw_qty

        # Create a shared qty manager between URLs for URL groups
        qty = manager.QtyManager(int_qty) if len(url_group) > 1 else QtyManager(int_qty)

        urls += url_group
        qtys += [qty] * len(url_group)

    with loading("Checking URLs"):
        titles = list(url_utils.get_url_titles())

    if len(titles) < len(urls):
        sys.exit(1)
    elif len(titles) > len(urls):
        Colors.print(
            "Something went wrong!",
            "Please report the issue to https://github.com/LeonShams/BestBuyBulletBot/issues.",
            "Feel free to copy and paste the following when opening an issue.",
            properties=["fail", "bold"],
        )
        print(
            "ERROR ENCOUNTERED DURING EXECUTION: More titles than URLs!",
            f"Raw URLs: {raw_urls}",
            f"URLs: {urls}",
            f"Titles: {titles}",
            sep="\n",
        )
        sys.exit(1)

    if AUTO_CHECKOUT:
        email, password, cvv = user_data.get_creds()
        if not (email or password or cvv):
            Colors.warn(
                "\nCheckout credentials have not been set. Run `3b-bot set-creds` to add the necessary information."
            )

            if not suppress_warnings:
                if yes_or_no("Would you like to set your checkout credentials (y/n): "):
                    user_data.set_creds()
                    email, password, cvv = user_data.get_creds()

        print()
        login_cookies_list, predicted_prices = browser_startup(
            headless, email, password, urls, verify_account, skip_verification
        )
        headers = {}
    else:
        email, password, cvv = "", "", ""
        login_cookies_list, predicted_prices = ((None for _ in urls) for i in range(2))

        headers = {
            "accept": "*/*",
            "accept-encoding": "gzip, deflate, br",
            "accept-language": "en-US,en;q=0.9,zh-CN;q=0.8,zh;q=0.7",
            "sec-ch-ua-mobile": "?0",
            "sec-fetch-dest": "script",
            "sec-fetch-mode": "no-cors",
            "sec-fetch-site": "same-origin",
            "user-agent": get_user_agent(),
        }

    Colors.print("Availability tracking has started!", properties=["success"])
    if fast:
        Colors.print("Fast tracking enabled!", properties=["blue"])
    print()

    # Create remaining shared objects
    thread_lock = manager.ThreadLock()
    paused = manager.PauseEvent()
    pbar = manager.IndefeniteProgressBar()

    # Start process for each URL
    with Pool(len(urls), set_priority, [fast]) as p:
        p.starmap(
            track,
            [
                [
                    title,
                    url,
                    qty,
                    headless,
                    login_cookies,
                    password,
                    cvv,
                    thread_lock,
                    paused,
                    pbar,
                    pred_price,
                    money_manager,
                    headers,
                ]
                for title, url, qty, login_cookies, pred_price in zip(
                    titles, urls, qtys, login_cookies_list, predicted_prices
                )
            ],
        )

    pbar.close()
    print("\nAll processes have finished.")



import os
from threading import Event, Thread

from playsound import playsound

path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "notification.wav")
playing = Event()


def _repeat():
    while True:
        playing.wait()
        playsound(path)


repeat_loop = Thread(target=_repeat, daemon=True)
repeat_loop.start()


def play(block=False):
    playsound(path, block)


def start():
    playing.set()


def stop():
    playing.clear()



import logging
import sys

import clipboard
import requests
from selenium.common.exceptions import NoSuchWindowException, TimeoutException
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait

from best_buy_bullet_bot.data.browser_login import (
    cookies_available,
    load_cookies,
    save_cookies,
)
from best_buy_bullet_bot.data.setting_utils import (
    DRIVER_NAMES,
    change_browser,
    get_settings,
    is_installed,
    update_setting,
)
from best_buy_bullet_bot.utils import Colors, loading

SETTINGS = get_settings()
TAX = SETTINGS["tax"]
BROWSER_NAME = SETTINGS["browser"]
DRIVER_WRAPPER = DRIVER_NAMES[BROWSER_NAME]
MAC = sys.platform == "darwin"
USER_TAKEOVER = 20 * 60  # 20 min for user to takeover if the bot gets stuck

logging.disable(logging.WARNING)
try:
    DRIVER_PATH = DRIVER_WRAPPER.manager().install()
except ValueError:
    if not is_installed(SETTINGS["browser"]):
        Colors.print(
            f"{SETTINGS['browser'].title()} is not installed on your computer.",
            properties=["fail"],
        )
        change_browser()


def _get_options(headless):
    options = DRIVER_WRAPPER.options()
    options.page_load_strategy = "none"
    options.add_argument("--proxy-server='direct://'")
    options.add_argument("--proxy-bypass-list=*")

    if headless:
        options.add_argument("--headless")

    # Suppress "DevTools listening on ws:..." message
    if BROWSER_NAME == "chrome":
        options.add_experimental_option("excludeSwitches", ["enable-logging"])

    return options


PREBUILT_OPTIONS = [_get_options(False), _get_options(True)]


def get_user_agent():
    driver = DRIVER_WRAPPER.driver(
        executable_path=DRIVER_PATH, options=PREBUILT_OPTIONS[True]
    )
    user_agent = driver.execute_script("return navigator.userAgent")
    driver.quit()
    return user_agent


def money2float(money):
    return float(money[1:].replace(",", ""))


def fast_text(text):
    clipboard.copy(text)
    return (Keys.COMMAND if MAC else Keys.CONTROL) + "v"


account_page_url = "https://www.bestbuy.com/site/customer/myaccount"
billing_url = "https://www.bestbuy.com/profile/c/billinginfo/cc"


def terminate(driver):
    driver.quit()
    sys.exit(1)


def _login(driver, wait, headless, email, password, cookies_set):
    branch = wait.until(
        EC.presence_of_element_located(
            (By.CSS_SELECTOR, "#ca-remember-me, .shop-search-bar")
        )
    )

    if branch.get_attribute("class") == "shop-search-bar":
        # We are already logged in
        return

    # Click "Keep me signed in" button
    branch.click()

    if not cookies_set:
        # Fill in email box
        driver.find_element_by_id("fld-e").send_keys(fast_text(email))

    # Fill in password box
    driver.find_element_by_id("fld-p1").send_keys(fast_text(password))

    # Click the submit button
    driver.find_element_by_css_selector(
        ".btn.btn-secondary.btn-lg.btn-block.c-button-icon.c-button-icon-leading.cia-form__controls__submit"
    ).click()

    # Check for error or redirect
    branch = wait.until(
        EC.presence_of_element_located(
            (
                By.CSS_SELECTOR,
                ".shop-search-bar, "  # We got redirected to the account page
                ".cia-cancel, "  # Skippable verification page
                ".c-alert.c-alert-level-error, "  # Error popup message
                "#fld-e-text, "  # Invalid email address
                "#fld-p1-text",  # Invalid password
            )
        )
    )

    # If we hit an error
    if branch.get_attribute(
        "class"
    ) == "c-alert c-alert-level-error" or branch.get_attribute("id") in [
        "fld-e-text",
        "fld-p1-text",
    ]:
        if headless:
            # If headless raise error
            Colors.print(
                "Incorrect login info. Please correct the username or password.",
                properties=["fail", "bold"],
            )
            terminate(driver)
        else:
            # If headful ask the user to take over
            Colors.print(
                "Unable to login automatically. Please correct your credentials or enter the information manually.",
                properties=["fail"],
            )
            branch = wait.until(
                EC.presence_of_element_located(
                    (By.CSS_SELECTOR, ".shop-search-bar, .cia-cancel")
                )
            )

    # If we hit a skippable verification page
    if "cia-cancel" in branch.get_attribute("class"):
        # Redirect to the "my account" page
        driver.get(account_page_url)
        wait.until(EC.presence_of_element_located((By.CLASS_NAME, "shop-search-bar")))

    save_cookies(driver)


def check_cart(driver):
    with loading("Confirming cart is empty"):
        # Confirm that the cart is empty
        headers = {
            "Host": "www.bestbuy.com",
            "User-Agent": driver.execute_script("return navigator.userAgent"),
            "Accept": "*/*",
            "Accept-Language": "en-US,en;q=0.5",
            "Accept-Encoding": "gzip, deflate, br",
            "Referer": "https://www.bestbuy.com/site/customer/myaccount",
            "X-CLIENT-ID": "browse",
            "X-REQUEST-ID": "global-header-cart-count",
            "Connection": "keep-alive",
            "Cookie": driver.execute_script("return document.cookie"),
            "Sec-Fetch-Dest": "empty",
            "Sec-Fetch-Mode": "cors",
            "Sec-Fetch-Site": "same-origin",
            "Cache-Control": "max-age=0",
        }

        # Make a get request to the Best Buy API to get the number of items in the cart
        response = requests.get(
            "https://www.bestbuy.com/basket/v1/basketCount", headers=headers
        )
        items = response.json()["count"]

        if items != 0:
            Colors.print(
                "Too many items in the cart. Please empty your cart before starting the bot.",
                properties=["fail", "bold"],
            )
            terminate(driver)


def perform_account_verification(driver, wait):
    with loading("Verifying account setup"):
        # Check that a shipping address has been set
        shipping_address = wait.until(
            EC.presence_of_element_located(
                (
                    By.CSS_SELECTOR,
                    "div.account-setting-block-container:nth-child(1) > div:nth-child(2) > a:last-child",
                )
            )
        )
        if shipping_address.get_attribute("class") == "":
            Colors.print(
                "Shipping address has not been set. You can add a shipping address to \
                your account at https://www.bestbuy.com/profile/c/address/shipping/add.",
                properties=["fail", "bold"],
            )
            terminate(driver)

        # Confirm that a default payment method has been created
        driver.get(billing_url)
        payment_method_list = wait.until(
            EC.presence_of_element_located(
                (
                    By.CSS_SELECTOR,
                    ".pf-credit-card-list__content-spacer > ul.pf-credit-card-list__credit-card-list",
                )
            )
        )

        if payment_method_list.size["height"] == 0:
            Colors.print(
                f"A default payment method has not been created. Please create one at {billing_url}.",
                properties=["fail", "bold"],
            )
            terminate(driver)

    Colors.print("Account has passed all checks!", properties=["success"])


def collect_item_cookies(driver, wait, urls):
    login_cookies_list = []
    predicted_prices = []
    price_element = None

    with loading("Collecting cookies for each URL"):
        for url in urls:
            driver.get(url)

            if price_element is not None:
                wait.until(EC.staleness_of(price_element))

            price_element = wait.until(
                EC.presence_of_element_located(
                    (
                        By.CSS_SELECTOR,
                        ".pricing-price > div > div > div > .priceView-hero-price.priceView-customer-price, "
                        ".pricing-price > div > div > div > div > section > div > div > .priceView-hero-price.priceView-customer-price",
                    )
                )
            )
            item_price = price_element.text.split("\n")[0]

            predicted_prices.append(money2float(item_price) * (1 + TAX))
            login_cookies_list.append(driver.get_cookies())

    return login_cookies_list, predicted_prices


def _browser_startup(
    driver, headless, email, password, urls, verify_account, skip_verification
):
    wait = WebDriverWait(driver, USER_TAKEOVER)

    with loading("Logging in"):
        driver.get(account_page_url)
        # We will then get redirected to the sign in page

        # If we have logged in previously we can use the cookies from that session
        # to skip steps in the login process and prevent the system from detecting
        # a bunch of logins from the same account
        cookies_exist = cookies_available()
        if cookies_exist:
            if load_cookies(driver):
                driver.refresh()
            else:
                # An error occurred while adding the login cookies
                cookies_exist = False

        _login(driver, wait, headless, email, password, cookies_exist)

    check_cart(driver)

    if not skip_verification:
        if SETTINGS["account verification"] or verify_account:
            perform_account_verification(driver, wait)
            if not verify_account:
                print("This was a one time test and will not be performed again.\n")
            update_setting("account verification", False)

    item_cookies = collect_item_cookies(driver, wait, urls)

    driver.quit()
    return item_cookies


def browser_startup(headless, *args, **kwargs):
    driver = DRIVER_WRAPPER.driver(
        executable_path=DRIVER_PATH, options=PREBUILT_OPTIONS[headless]
    )

    try:
        return _browser_startup(driver, headless, *args, **kwargs)

    # Timed out while trying to locate element
    except TimeoutException:
        Colors.print(
            "Browser window has timed out. Closing bot.", properties=["fail", "bold"]
        )
        terminate(driver)

    # User has closed the browser window
    except NoSuchWindowException:
        terminate(driver)


def _purchase(
    driver,
    title,
    password,
    cvv,
    money_manager,
):
    # Go to the checkout page
    driver.get("https://www.bestbuy.com/checkout/r/fast-track")
    wait = WebDriverWait(driver, USER_TAKEOVER)

    # Get to the CVV page
    while True:
        branch = wait.until(
            EC.element_to_be_clickable(
                (
                    By.CSS_SELECTOR,
                    "#credit-card-cvv, "  # Place order page
                    ".button--continue > button.btn.btn-lg.btn-block.btn-secondary, "  # Continue to payment info page
                    ".checkout-buttons__checkout > .btn.btn-lg.btn-block.btn-primary",  # We got redirected to the cart
                )
            )
        )

        # If we got redirected to the cart
        if branch.get_attribute("class") == "btn btn-lg btn-block btn-primary":
            # Click "proceed to checkout" button
            branch.click()
            branch = wait.until(
                EC.element_to_be_clickable(
                    (
                        By.CSS_SELECTOR,
                        "#credit-card-cvv, "  # Place order page
                        "#cvv, "  # Review and place order page
                        ".button--continue > button.btn.btn-lg.btn-block.btn-secondary, "  # Continue to place order page (page before place order page)
                        "#fld-p1",  # Sign in (only requires password)
                    )
                )
            )

            # If it wants to confirm our password
            if branch.get_attribute("class").strip() == "tb-input":
                branch.send_keys(fast_text(password))
                driver.find_element_by_css_selector(
                    ".btn.btn-secondary.btn-lg.btn-block.c-button-icon.c-button-icon-leading.cia-form__controls__submit"
                ).click()  # Click sign in button

                # We will loop back around and handle what comes next
            else:
                break
        else:
            break

    # Select the CVV text box
    if branch.get_attribute("class") == "btn btn-lg btn-block btn-secondary":
        branch.click()
        cvv_box = wait.until(
            EC.element_to_be_clickable(
                (
                    By.CSS_SELECTOR,
                    "#credit-card-cvv, #cvv",
                )
            )
        )
    else:
        cvv_box = branch
    cvv_box.send_keys(fast_text(cvv))

    # Locate and parse the grand total text
    grand_total = money2float(
        driver.find_element_by_css_selector(
            ".order-summary__total > .order-summary__price > .cash-money"
        ).text
    )

    # Make sure we have sufficient funds for the purchase
    if money_manager.check_funds(grand_total):
        # Click place order button
        driver.find_element_by_css_selector(
            ".btn.btn-lg.btn-block.btn-primary, .btn.btn-lg.btn-block.btn-primary.button__fast-track"
        ).click()

        # Deduct grand total from available funds
        money_manager.make_purchase(grand_total)

        Colors.print(
            f"Successfully purchased {title}. The item was a grand total of ${grand_total:,.2f} leaving you with ${money_manager.get_funds():,.2f} of available funds.",
            properties=["success", "bold"],
        )
        return True
    else:
        Colors.print(
            f"Insuffient funds to purchase {title} which costs a grand total of ${grand_total:,.2f} while you only have ${money_manager.get_funds():,.2f} of available funds.",
            properties=["fail"],
        )
    return False


def purchase(
    url, login_cookies, headless, headless_driver, headless_wait, *args, **kwargs
):
    if not headless:
        # Create a new visible driver for the checkout process
        driver = DRIVER_WRAPPER.driver(
            executable_path=DRIVER_PATH, options=PREBUILT_OPTIONS[False]
        )
        driver.get(url)
        for cookie in login_cookies:
            driver.add_cookie(cookie)
    else:
        # Use the old headless driver so we don't have to create a new one
        driver = headless_driver

    # Have the existing headless tracker driver click the add-to-cart button
    headless_wait.until(
        EC.element_to_be_clickable(
            (
                By.CSS_SELECTOR,
                ".fulfillment-add-to-cart-button > div > div > button",
            )
        )
    ).click()

    try:
        return _purchase(driver, *args, **kwargs)
    except TimeoutException:
        Colors.print(
            "3B Bot got stuck and nobody took over. Tracking will resume.",
            properties=["fail"],
        )
    except NoSuchWindowException:
        driver.quit()

    return False



import sys
import time
from collections import deque
from datetime import timedelta

from rich import get_console
from rich.progress import BarColumn, Progress, ProgressColumn, SpinnerColumn, TextColumn


class TimeRemainingColumn(ProgressColumn):
    """Renders estimated time remaining."""

    # Only refresh twice a second to prevent jitter
    max_refresh = 0.5

    def __init__(self, *args, **kwargs):
        self.start_time = time.time()
        super().__init__(*args, **kwargs)

    def render(self, *args, **kwargs):
        delta = timedelta(seconds=int(time.time() - self.start_time))
        return str(delta)


class IterationsPerSecond:
    def format(self, task):
        if "times" in dir(task) and len(task.times):
            speed = len(task.times) / task.times[-1]
            return f"{speed:.2f}it/s"
        return "0.00it/s"


class IndefeniteProgressBar:
    def __init__(self):
        with get_console() as console:
            self.pbar = Progress(
                SpinnerColumn(style=""),
                TextColumn("{task.completed}it"),
                BarColumn(console.width),
                TextColumn(IterationsPerSecond()),
                TimeRemainingColumn(),
                console=console,
                expand=True,
            )

        self.pbar.start()
        self.pbar.add_task(None, start=False)
        self.pbar.tasks[0].times = deque(maxlen=100)
        self.start_time = time.time()

    def print(self, *args, sep=" ", end="\n"):
        msg = sep.join(map(str, args))
        sys.stdout.writelines(msg + end)

    def update(self):
        task = self.pbar.tasks[0]
        task.completed += 1
        task.times.append(time.time() - self.start_time)

    def close(self):
        self.pbar.stop()



import argparse
import warnings

from best_buy_bullet_bot.utils import count_cores
from best_buy_bullet_bot.version import __version__


class NoAction(argparse.Action):
    """Makes argument do nothing.

    This is useful if we want an argument to show up in the
    help menu, but remain uncallable.
    """

    def __init__(self, **kwargs):
        kwargs.setdefault("default", argparse.SUPPRESS)
        kwargs.setdefault("nargs", 0)
        super().__init__(**kwargs)

    def __call__(self, *args):
        pass


class FuncKwargs(dict):
    """Only passes flags to a specified function."""

    def __init__(self, args):
        self.args = args
        super().__init__()

    def add_flag(self, flag_name, cmd_name):
        flag = getattr(self.args, flag_name)
        flag and self.args.cmd == cmd_name and self.update({flag_name: flag})


class ImportWrapper:
    """Only imports the function that the user selects."""

    def __init__(self, file):
        self.file = file

    def __getattribute__(self, name):
        if name == "file":
            return super().__getattribute__("file")

        def call_func(*args, **kwargs):
            imported_file = __import__(self.file, fromlist=[""])
            return getattr(imported_file, name)(*args, **kwargs)

        return call_func


# This is done to prevent unnecessary imports and more importantly
# prevent a bunch of warnings from setting_utils when imported
tracker = ImportWrapper("best_buy_bullet_bot.tracker")
setting_utils = ImportWrapper("best_buy_bullet_bot.data.setting_utils")
url_utils = ImportWrapper("best_buy_bullet_bot.data.url_utils")
user_data = ImportWrapper("best_buy_bullet_bot.data.user_data")
browser_login = ImportWrapper("best_buy_bullet_bot.data.browser_login")

OPS = {
    "start": [tracker.start, "Start tracking the currently set URLs."],
    "view-urls": [url_utils.view_urls, "View list of tracked URLs."],
    "add-url": [url_utils.add_url, "Add URL to tracking list."],
    "add-url-group": [
        url_utils.add_url_group,
        "Add multiple URLs and set a quantity for all of them as a whole instead of individually.",
    ],
    "remove-url": [
        url_utils.remove_url,
        "Remove a URL from the list of tracked URLs.",
    ],
    "test-urls": [
        url_utils.test_urls,
        "Tests to make sure all URLs can be tracked. This is also run on startup.",
    ],
    "clear-urls": [url_utils.clear_urls, "Remove all tracked URLs."],
    "view-settings": [setting_utils.view_settings, "View current settings."],
    "set-funds": [
        setting_utils.set_funds,
        "Set how much money the bot is allowed to spend.",
    ],
    "set-tax": [setting_utils.set_tax, "Set the sales tax rate for your state."],
    "toggle-auto-checkout": [
        setting_utils.toggle_auto_checkout,
        "Enable/disable auto checkout.",
    ],
    "change-browser": [
        setting_utils.change_browser,
        "Pick the browser to be used during tracking and auto-checkout (only applies if auto-checkout is enabled). \
            Firefox is the default and recommended browser.",
    ],
    "test-sound": [setting_utils.test_sound, "Play sound sample."],
    "set-sound-mode": [
        setting_utils.set_sound_mode,
        "Choose whether you want sound to be completely disabled, play once on item restock, or play repeatedly on item restock.",
    ],
    "set-threads": [
        setting_utils.set_threads,
        "Select the number of threads to allocate to tracking each URL.",
    ],
    "count-cores": [
        count_cores,
        "Print how many CPU cores you have and how many threads each core has.",
    ],
    "reset-settings": [
        setting_utils.reset_settings,
        "Reset setting to the defaults.",
    ],
    "view-creds": [
        user_data.print_creds,
        "View your Best Buy login credentials (email, password, cvv).",
    ],
    "set-creds": [
        user_data.set_creds,
        "Set your Best Buy login credentials (email, password, cvv).",
    ],
    "clear-creds": [
        user_data.clear_creds,
        "Reset your Best Buy login credentials (email, password, cvv). Also offers the option to reset your access password.",
    ],
}


def run_command():
    parser = argparse.ArgumentParser(
        prog="3b-bot",
        description="Setup and control your Best Buy bot.",
        epilog="Good luck :)",
    )

    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version="%(prog)s " + __version__,
        help="show 3B Bot version number",
    )

    parser.add_argument(
        "cmd",
        default="start",
        const="start",
        nargs="?",
        choices=OPS.keys(),
        help="Performs a specified operation.",
        metavar="command",
        type=str.lower,
    )
    group = parser.add_argument_group(title="Available commands")

    for name, [func, help_msg] in OPS.items():
        group.add_argument(name, help=help_msg, action=NoAction)

    parser.add_argument(
        "-w", "--suppress-warnings", action="store_true", help="suppress warnings"
    )

    """
    EASTER EGG: Thank you for reading the source code!
    To run the bot with a higher priority level and achieve better performance complete the following.

    If using Firefox, complete the following before moving on to the next step:
    WINDOWS: Open a Command Prompt window with "Run as administrator" https://www.educative.io/edpresso/how-to-run-cmd-as-an-administrator
    MAC: Enter the command `su` in your terminal to gain root privileges. Beware your settings may be different in the root session, but you can always return to a normal session with the `exit` command.

    Then regardless of your browser:
    Run `3b-bot --fast` in your shell.
    """
    parser.add_argument("--fast", action="store_true", help=argparse.SUPPRESS)
    parser.add_argument(
        "--headless", action="store_true", help="hide the browser during auto checkout"
    )
    parser.add_argument(
        "--verify-account",
        action="store_true",
        help="confirm that the account is setup properly (automatically performed on first run)",
    )
    parser.add_argument(
        "--skip-verification",
        action="store_true",
        help="skip checks on first run that make sure account is setup properly.",
    )
    parser.add_argument(
        "--force-login",
        action="store_true",
        help="force browser to go through traditional login process as opposed to using cookies to skip steps",
    )

    args = parser.parse_args()
    func_kwargs = FuncKwargs(args)

    func_kwargs.add_flag("fast", "start")
    func_kwargs.add_flag("headless", "start")
    func_kwargs.add_flag("verify_account", "start")
    func_kwargs.add_flag("skip_verification", "start")

    if args.suppress_warnings:
        warnings.filterwarnings("ignore")
    else:
        # Just ignore the depreciation warnings
        warnings.filterwarnings("ignore", category=DeprecationWarning)

    if args.force_login:
        browser_login.delete_cookies()

    # Run command
    OPS[args.cmd][0](**func_kwargs)



from pathlib import Path

from setuptools import find_packages, setup

# Get __version__ from best_buy_bullet_bot/version.py
exec(
    compile(
        open("best_buy_bullet_bot/version.py").read(),
        "best_buy_bullet_bot/version.py",
        "exec",
    )
)

setup(
    name="3b-bot",
    version=__version__,
    author="Leon Shams-Schaal",
    description="Quickly purchase items from Best Buy the moment they restock.",
    long_description=Path("readme.md").read_text(encoding="utf-8"),
    long_description_content_type="text/markdown",
    url="https://github.com/LeonShams/BestBuyBulletBot",
    project_urls={
        "Documentation": "https://github.com/LeonShams/BestBuyBulletBot/wiki",
        "Bug Tracker": "https://github.com/LeonShams/BestBuyBulletBot/issues",
    },
    packages=find_packages(),
    python_requires=">=3.6",
    install_requires=Path("requirements.txt").read_text().splitlines(),
    entry_points={"console_scripts": ["3b-bot=best_buy_bullet_bot.__main__:main"]},
    include_package_data=True,
    zip_safe=False,
    license="MIT",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: End Users/Desktop",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Environment :: Console",
    ],
    keywords="bestbuy bot bestbuybot bestbuybulletbot 3bbot 3b-bot bestbuystock \
        bestbuyrestock bestbuytracker stocktracker bestbuystocktracker \
        autocheckout bestbuyautocheckout nvidiabot gpubot nvidiagpubot \
        3060bot 3060tibot 3070bot 3070tibot 3080bot 3080tibot 3090bot \
        ps5bot nvidiatracker gputracker nvidiagputracker 3060tracker \
        3060titracker 3070tracker 3070titracker 3080tracker 3080titracker \
        3090tracker ps5tracker",
)



import os
import sys
from getpass import getpass

from keyring.errors import PasswordDeleteError
from keyring.util import properties
from keyrings.cryptfile.cryptfile import CryptFileKeyring
from keyrings.cryptfile.file_base import FileBacked

from best_buy_bullet_bot.data import SHARED_DIR
from best_buy_bullet_bot.utils import Colors, yes_or_no

EMAIL_VAR = "BB_EMAIL"
PASS_VAR = "BB_PASS"
CVV_VAR = "BB_CVV"


@properties.NonDataProperty
def file_path(self):
    return os.path.join(SHARED_DIR, self.filename)


FileBacked.file_path = file_path

SERVICE_ID = "3B_BOT"
KR = CryptFileKeyring()


def set_access_pass(access_pass):
    KR._keyring_key = access_pass


def authenticate():
    attempts = 0
    while True:
        try:
            KR.keyring_key
            return
        except ValueError:
            if str(KR._keyring_key).strip() == "":
                sys.exit()

            attempts += 1
            if attempts >= 3:
                print("Too many attempts, please try again later.")
                sys.exit()

            print("Sorry, try again.")
            KR._keyring_key = None


def _get_cred(name, default_value):
    cred = KR.get_password(SERVICE_ID, name)
    return cred if cred is not None else default_value


def get_creds(default_value=""):
    authenticate()
    return [_get_cred(var, default_value) for var in [EMAIL_VAR, PASS_VAR, CVV_VAR]]


def _get_input(prompt):
    while True:
        value = input(prompt)

        if yes_or_no("Continue (y/n): "):
            return value


def set_creds():
    authenticate()
    KR.set_password(SERVICE_ID, EMAIL_VAR, _get_input("Email: "))

    print()
    while True:
        password = getpass("Best Buy password: ")
        confirm_pass = getpass("Confirm password: ")
        if password == confirm_pass:
            break
        print("Passwords didn't match! Try again.")
    KR.set_password(SERVICE_ID, PASS_VAR, password)

    print()
    KR.set_password(SERVICE_ID, CVV_VAR, _get_input("CVV: "))
    Colors.print("Successfully updated credentials!", properties=["success"])


def print_creds():
    email, password, cvv = get_creds(Colors.str("EMPTY", ["fail"]))
    print("Email:", email)
    print("Password:", password)
    print("CVV:", cvv)


def clear_creds():
    for name in [EMAIL_VAR, PASS_VAR, CVV_VAR]:
        try:
            KR.delete_password(SERVICE_ID, name)
        except PasswordDeleteError:
            pass
    Colors.print("Credentials cleared!\n", properties=["success", "bold"])

    # Check if user wants to reset their password
    if yes_or_no("Would you like to reset your password (y/n): "):
        os.remove(KR.file_path)
        KR.keyring_key
